@@ -3,8 +3,8 @@ use std::fmt;
 use std::io;
 use std::result;
 
-use fst;
 use clap;
+use fst;
 use ucd_parse;
 use ucd_trie;
 
@@ -27,14 +27,6 @@ impl Error {
 }
 
 impl error::Error for Error {
-    fn description(&self) -> &str  {
-        match *self {
-            Error::Io(ref err) => err.description(),
-            Error::Clap(ref err) => err.description(),
-            Error::Other(ref msg) => msg,
-        }
-    }
-
     fn source(&self) -> Option<&(dyn error::Error + 'static)> {
         match *self {
             Error::Io(ref err) => Some(err),
@@ -45,7 +37,7 @@ impl error::Error for Error {
 }
 
 impl fmt::Display for Error {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match *self {
             Error::Io(ref err) => err.fmt(f),
             Error::Clap(ref err) => err.fmt(f),
@@ -83,9 +75,3 @@ impl From<ucd_trie::Error> for Error {
         Error::Other(err.to_string())
     }
 }
-
-impl From<regex_automata::Error> for Error {
-    fn from(err: regex_automata::Error) -> Error {
-        Error::Other(err.to_string())
-    }
-}
