@@ -1,12 +1,12 @@
 use std::collections::{BTreeMap, BTreeSet};
 
-use ucd_parse::{self, UnicodeDataExpander};
+use ucd_parse::{self, UnicodeData, UnicodeDataExpander};
 
-use args::ArgMatches;
-use error::Result;
-use util::{PropertyValues, print_property_values};
+use crate::args::ArgMatches;
+use crate::error::Result;
+use crate::util::{print_property_values, PropertyValues};
 
-pub fn command(args: ArgMatches) -> Result<()> {
+pub fn command(args: ArgMatches<'_>) -> Result<()> {
     let dir = args.ucd_dir()?;
     let propvals = PropertyValues::from_ucd_dir(&dir)?;
     let filter = args.filter(|name| propvals.canonical("gc", name))?;
@@ -18,37 +18,13 @@ pub fn command(args: ArgMatches) -> Result<()> {
         return print_property_values(&propvals, "General_Category");
     }
 
-    // Expand all of our UnicodeData rows. This results in one big list of
-    // all assigned codepoints.
-    let rows: Vec<_> = UnicodeDataExpander::new(unexpanded).collect();
+    let mut bycat = expand_into_categories(unexpanded, &propvals)?;
 
-    // Collect each general category into an ordered set.
-    let mut bycat: BTreeMap<String, BTreeSet<u32>> = BTreeMap::new();
-    let mut assigned = BTreeSet::new();
-    for row in rows {
-        assigned.insert(row.codepoint.value());
-        let gc = propvals
-            .canonical("gc", &row.general_category)?
-            .to_string();
-        bycat.entry(gc)
-            .or_insert(BTreeSet::new())
-            .insert(row.codepoint.value());
-    }
-    // As a special case, collect all unassigned codepoints.
-    let unassigned_name = propvals
-        .canonical("gc", "unassigned")?
-        .to_string();
-    bycat.insert(unassigned_name.clone(), BTreeSet::new());
-    for cp in 0..(0x10FFFF + 1) {
-        if !assigned.contains(&cp) {
-            bycat.get_mut(&unassigned_name).unwrap().insert(cp);
-        }
-    }
     // As another special case, collect all "related" groups of categories.
     // But don't do this when printing an enumeration, because in an
     // enumeration each codepoint should belong to exactly one category, which
     // is not true if we include related categories.
-    if !args.is_present("enum") {
+    if !args.is_present("enum") && !args.is_present("rust-enum") {
         for (name, set) in related(&propvals, &bycat) {
             if filter.contains(&name) {
                 bycat.insert(name, set);
@@ -64,6 +40,11 @@ pub fn command(args: ArgMatches) -> Result<()> {
     let mut wtr = args.writer("general_category")?;
     if args.is_present("enum") {
         wtr.ranges_to_enum(args.name(), &bycat)?;
+    } else if args.is_present("rust-enum") {
+        let variants = bycat.keys().map(String::as_str).collect::<Vec<_>>();
+        wtr.ranges_to_rust_enum(args.name(), &variants, &bycat)?;
+    } else if args.is_present("combined") {
+        wtr.ranges_to_combined(args.name(), &bycat)?;
     } else {
         wtr.names(bycat.keys().filter(|n| filter.contains(n)))?;
         for (name, set) in bycat {
@@ -74,6 +55,38 @@ pub fn command(args: ArgMatches) -> Result<()> {
     Ok(())
 }
 
+/// Expand a list of UnicodeData rows and group by category.
+pub fn expand_into_categories(
+    unexpanded: Vec<UnicodeData>,
+    propvals: &PropertyValues,
+) -> Result<BTreeMap<String, BTreeSet<u32>>> {
+    // Expand all of our UnicodeData rows. This results in one big list of
+    // all assigned codepoints.
+    let rows: Vec<_> = UnicodeDataExpander::new(unexpanded).collect();
+
+    // Collect each general category into an ordered set.
+    let mut bycat: BTreeMap<String, BTreeSet<u32>> = BTreeMap::new();
+    let mut assigned = BTreeSet::new();
+    for row in rows {
+        assigned.insert(row.codepoint.value());
+        let gc = propvals.canonical("gc", &row.general_category)?.to_string();
+        bycat
+            .entry(gc)
+            .or_insert(BTreeSet::new())
+            .insert(row.codepoint.value());
+    }
+    // As a special case, collect all unassigned codepoints.
+    let unassigned_name = propvals.canonical("gc", "unassigned")?.to_string();
+    bycat.insert(unassigned_name.clone(), BTreeSet::new());
+    for cp in 0..=0x10FFFF {
+        if !assigned.contains(&cp) {
+            bycat.get_mut(&unassigned_name).unwrap().insert(cp);
+        }
+    }
+
+    Ok(bycat)
+}
+
 /// Related returns a set of sets of codepoints corresponding to the "related"
 /// groups of categories defined by Table 12 in UAX#44 S5.7.1.
 ///
@@ -106,9 +119,18 @@ fn related_categories(
         (c("Letter"), vec![c("lu"), c("ll"), c("lt"), c("lm"), c("lo")]),
         (c("Mark"), vec![c("mn"), c("mc"), c("me")]),
         (c("Number"), vec![c("nd"), c("nl"), c("no")]),
-        (c("Punctuation"), vec![
-            c("pc"), c("pd"), c("ps"), c("pe"), c("pi"), c("pf"), c("po"),
-        ]),
+        (
+            c("Punctuation"),
+            vec![
+                c("pc"),
+                c("pd"),
+                c("ps"),
+                c("pe"),
+                c("pi"),
+                c("pf"),
+                c("po"),
+            ],
+        ),
         (c("Symbol"), vec![c("sm"), c("sc"), c("sk"), c("so")]),
         (c("Separator"), vec![c("zs"), c("zl"), c("zp")]),
         (c("Other"), vec![c("cc"), c("cf"), c("cs"), c("co"), c("cn")]),
