@@ -2,14 +2,15 @@ use std::collections::{BTreeMap, BTreeSet};
 use std::path::Path;
 
 use ucd_parse::{
-    self, CoreProperty, EmojiProperty, Property, UnicodeDataExpander,
+    self, CoreProperty, EmojiProperty, Property, UcdFileByCodepoint,
+    UnicodeData, UnicodeDataExpander,
 };
 
-use args::ArgMatches;
-use error::Result;
-use util::{PropertyNames, PropertyValues};
+use crate::args::ArgMatches;
+use crate::error::Result;
+use crate::util::{PropertyNames, PropertyValues};
 
-pub fn command(args: ArgMatches) -> Result<()> {
+pub fn command(args: ArgMatches<'_>) -> Result<()> {
     let dir = args.ucd_dir()?;
     let by_name = parse_properties(&dir)?;
     let properties = PropertyNames::from_ucd_dir(&dir)?;
@@ -31,7 +32,7 @@ pub fn command(args: ArgMatches) -> Result<()> {
     Ok(())
 }
 
-pub fn command_perl_word(args: ArgMatches) -> Result<()> {
+pub fn command_perl_word(args: ArgMatches<'_>) -> Result<()> {
     let dir = args.ucd_dir()?;
     let props = parse_properties(&dir)?;
     let gencats = parse_general_categories(&dir)?;
@@ -79,24 +80,34 @@ fn parse_properties<P: AsRef<Path>>(
             .extend(x.codepoints.into_iter().map(|c| c.value()));
     }
 
+    // Add Bidi_Mirrored
+    let unicode_data: Vec<UnicodeData> = ucd_parse::parse(&ucd_dir)?;
+    let bidi_mirrored =
+        unicode_data.iter().fold(BTreeSet::new(), |mut set, x| {
+            if x.bidi_mirrored {
+                set.extend(x.codepoints().into_iter().map(|c| c.value()))
+            }
+            set
+        });
+    by_name.insert("Bidi_Mirrored".to_string(), bidi_mirrored);
+
     // Since emoji-data.txt isn't parse of the normal UCD download, don't
     // die if it doesn't exist. But emit a helpful warning message.
     let emoji_prop: Vec<EmojiProperty> = match ucd_parse::parse(&ucd_dir) {
         Ok(props) => props,
-        Err(err) => {
-            match *err.kind() {
-                ucd_parse::ErrorKind::Io(_) => {
-                    eprintln!(
-                        "{}. skipping emoji properties. \
-                        emoji-data.txt can be downloaded from \
-                        https://unicode.org/Public/emoji/",
-                        err,
-                    );
-                    vec![]
-                }
-                _ => return Err(From::from(err)),
+        Err(err) => match *err.kind() {
+            ucd_parse::ErrorKind::Io(_) => {
+                eprintln!(
+                    "{}. skipping emoji properties. \
+                     emoji-data.txt is included in UCD 13.0.0 and newer, and \
+                     can be downloaded from https://unicode.org/Public/emoji/ \
+                     for older releases.",
+                    err,
+                );
+                vec![]
             }
-        }
+            _ => return Err(From::from(err)),
+        },
     };
     for x in &emoji_prop {
         by_name
@@ -119,10 +130,9 @@ fn parse_general_categories<P: AsRef<Path>>(
     // Collect each general category into an ordered set.
     let mut bycat: BTreeMap<String, BTreeSet<u32>> = BTreeMap::new();
     for row in rows {
-        let gc = propvals
-            .canonical("gc", &row.general_category)?
-            .to_string();
-        bycat.entry(gc)
+        let gc = propvals.canonical("gc", &row.general_category)?.to_string();
+        bycat
+            .entry(gc)
             .or_insert(BTreeSet::new())
             .insert(row.codepoint.value());
     }
