@@ -5,7 +5,7 @@ use std::str;
 use ucd_parse::{PropertyAlias, PropertyValueAlias};
 use ucd_util;
 
-use error::Result;
+use crate::error::Result;
 
 /// Filter is an include/exclude filter of strings specified on the command
 /// line via --include and --exclude flags.
@@ -132,9 +132,7 @@ impl PropertyValues {
                 value
             };
 
-            let inner_map = outer_map
-                .entry(prop)
-                .or_insert(BTreeMap::new());
+            let inner_map = outer_map.entry(prop).or_insert(BTreeMap::new());
             if let Some(n) = a.numeric {
                 inner_map.insert(make_key(n.to_string()), canon.clone());
             }
@@ -172,24 +170,22 @@ impl PropertyValues {
     ///
     /// Note that this does not apply to "string" or "miscellaneous" properties
     /// such as `Name` or `Case_Folding`.
-    pub fn canonical(
-        &self,
-        property: &str,
-        value: &str,
-    ) -> Result<String> {
+    pub fn canonical(&self, property: &str, value: &str) -> Result<String> {
         let property = self.property.canonical(property)?;
         let mut value = value.to_string();
         ucd_util::symbolic_name_normalize(&mut value);
         match self.value.get(&*property).and_then(|m| m.get(&value)) {
             Some(v) => Ok(v.to_string()),
             None => err!(
-                "unrecognized property name/value: {:?}", (property, value)),
+                "unrecognized property name/value: {:?}",
+                (property, value)
+            ),
         }
     }
 }
 
 /// Convert an iterator of codepoints into a vec of sorted ranges.
-pub fn to_ranges<I: IntoIterator<Item=u32>>(it: I) -> Vec<(u32, u32)> {
+pub fn to_ranges<I: IntoIterator<Item = u32>>(it: I) -> Vec<(u32, u32)> {
     let mut codepoints: Vec<u32> = it.into_iter().collect();
     codepoints.sort();
     codepoints.dedup();
@@ -222,10 +218,12 @@ pub fn range_add(ranges: &mut Vec<(u32, u32)>, codepoint: u32) {
 /// ranges.
 ///
 /// This panics if the same codepoint is present multiple times.
-pub fn to_range_values<I>(it: I) -> Vec<(u32, u32, u64)>
-    where I: IntoIterator<Item=(u32, u64)>
+pub fn to_range_values<I, V>(it: I) -> Vec<(u32, u32, V)>
+where
+    I: IntoIterator<Item = (u32, V)>,
+    V: Ord,
 {
-    let mut codepoints: Vec<(u32, u64)> = it.into_iter().collect();
+    let mut codepoints: Vec<(u32, V)> = it.into_iter().collect();
     codepoints.sort();
     codepoints.dedup();
 
@@ -244,14 +242,14 @@ pub fn to_range_values<I>(it: I) -> Vec<(u32, u32, u64)>
 ///
 /// This panics if the given codepoint is already in the ranges or if a
 /// codepoint is given out of order.
-pub fn range_value_add(
-    ranges: &mut Vec<(u32, u32, u64)>,
+pub fn range_value_add<V: Eq>(
+    ranges: &mut Vec<(u32, u32, V)>,
     codepoint: u32,
-    value: u64,
+    value: V,
 ) {
-    if let Some(&mut (_, ref mut end, value2)) = ranges.last_mut() {
+    if let Some(&mut (_, ref mut end, ref value2)) = ranges.last_mut() {
         assert!(*end < codepoint);
-        if codepoint == *end + 1 && value == value2 {
+        if codepoint == *end + 1 && &value == value2 {
             *end = codepoint;
             return;
         }
